'From Squeak3.0 of 4 February 2001 [latest update: #3545] on 21 July 2001 at 1:37:28 pm'!
Magnitude subclass: #ComplexNumbers
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MathPack-ComplexNumber'!

!ComplexNumbers commentStamp: 'jpwj 7/19/2001 16:00' prior: 0!
"Abstract class for the two forms of complex numbers.
            Class               Form
        ComplexNumber       algebriac form   a+bi
        ComplexPolarForm    polar form r(cos (theta) + i sin(theta))
    All arithmetic operations MUST be done with class ComplexNumber.
    ComplexPolarForm was implemented to facilate finding powers and
    roots of a ComplexNumber.
    The definition of complex numbers does not support the
    inequality messages, <  >  <=  >=  min:, max:, between:and:
    of super class Magnitude.  All these messages answer, ^self invalidMessage
    Aliases
      Trigonometric form - polar form
      modulus - absolute value (r of polar form)
      amplitude - argument (theta of polar form)
    "!

ComplexNumbers subclass: #ComplexNumber
	instanceVariableNames: 'real imaginary '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MathPack-ComplexNumber'!

!ComplexNumber commentStamp: 'jpwj 7/19/2001 14:55' prior: 0!
ComplexNumber is the algebraic form of a complex number.
    i.e. a + bi where i is implied and equal to the square root of -1.
    !

ComplexNumbers subclass: #ComplexPolarForm
	instanceVariableNames: 'modulus amplitude '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MathPack-ComplexNumber'!

!ComplexPolarForm commentStamp: 'jpwj 7/19/2001 15:13' prior: 0!
"ComplexPolarForm implemented primarily for finding the
    power of a complex number (aComplexNumber raisedTo: aNumber)
    OR n nth-roots of a complex number
        (aComplexNumber complexRoots: anInteger).
    Operators *,+,-,/  change receiver to ComplexNumber.
    Other arithmetic operations implemented in class ComplexNumber.

    trigonometric form conjugate
        of r( cos theta + i sin theta) = r( cos (-theta) + i sin (-theta))
        where -theta equals -theta + 360 degrees.
    negative of trignometric form
        of r( cos theta + i sin theta) equals
        r( cos (theta + 180 degrees) + i sin (theta + 180 degrees))
        "!


!Object methodsFor: 'testing' stamp: 'jpwj 7/19/2001 15:22'!
isScalar
   "Answer false since any object not belonging to class Number
    or ComplexNumbers is not a scalar. "

    ^false! !


!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:34'!
< aNumber
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '< comparison not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:33'!
<= aNumber
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '<= comparison not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:34'!
= aNumber
   "Answer true if the real,imaginary parts of two complex numbers
    are equal, else answer false."
    | epsilon |
    epsilon := 1.0e-16.
    (aNumber isKindOf: ComplexNumbers)
        ifFalse:[ ^false ].
    ( (aNumber real - self real) abs < epsilon
        and: [ (aNumber imaginary - self imaginary) abs < epsilon])
        ifTrue:[ ^true ].
    ^false! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:35'!
> aNumber
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '> comparison not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:36'!
>= aNumber
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '>= comparison not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:37'!
between: min and: max
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '#between:and: not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:36'!
max: aMagnitude
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '#max:  not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'comparing' stamp: 'jpwj 7/19/2001 15:36'!
min: aMagnitude
   "Operation not defined for the field of ComplexNumber(s)"

    ^self error: '#min:  not defined for the field of ComplexNumber(s)'! !

!ComplexNumbers methodsFor: 'copying' stamp: 'jpwj 7/19/2001 15:30'!
shallowCopy
   "Answer the receiver."
    ^self! !


!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:02'!
conjugate
   "Answer the conjugate of the receiver"
    ^ComplexNumber
        real: real
        imaginary: imaginary negated! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:02'!
denominator
   "Answer the denominator of self which is 1."
    ^1! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:02'!
exp
   "Answer the exponential value of self."
    | r i |
    r := real exp.
    i := r * ( imaginary sin).
    r := r * ( imaginary cos).
    ^ComplexNumber real: r imaginary: i! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:03'!
forceToZero: ifLessThanEpsilon
   "Answer ONLY the adjusted real part when the adjusted imaginary part
    equals zero, ELSE ANSWER new instance of self with adjusted
    real & imaginary parts.  #forceToZero: answers zero if
    the absolute value is less than epsilon else the receiver. "
    | i |
    i := imaginary forceToZero: ifLessThanEpsilon.
    i = 0
      ifTrue:[ ^real forceToZero: ifLessThanEpsilon ]
      ifFalse:[ ^ComplexNumber
                    real: ( real forceToZero: ifLessThanEpsilon)
                    imaginary: i  ]! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:03'!
hash
   "Answer anInteger hash value for receiver"
    ^real hash + imaginary hash! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:03'!
imaginary
        "Answer imaginary part of self."
    ^imaginary! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:04'!
modulus
   "Answer the absolute value of the receiver. "
    ^(real pythagorean: imaginary)! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:05'!
numerator
   "Answer the numerator of self which is self."
    ^self! !

!ComplexNumber methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:09'!
real
   "Answer the real part of self."
    ^real! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 14:58'!
* aNumber
   "Answer the product of receiver * aNumber"
    ^ComplexNumber
        real: ((real * aNumber real) - (imaginary * aNumber imaginary))
        imaginary: ((real * aNumber imaginary) + (imaginary * aNumber real))
! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 14:58'!
+ aNumber
   "Answer the sum of receiver + aNumber"
    ^ComplexNumber
        real: (real + aNumber real)
        imaginary: (imaginary + aNumber imaginary)
! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 14:59'!
- aNumber
   "Answer the difference of receiver - aNumber"
    ^ComplexNumber
        real: (real - aNumber real)
        imaginary: (imaginary - aNumber imaginary)! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 14:59'!
/ aNumber
   "Answer the quotient of receiver / aNumber"
    | complex conjugate denominator numerator |
    complex := aNumber asComplexNumber.
    conjugate := aNumber conjugate.
    numerator := self * conjugate.
    "ingore roundoff by using real part, imaginary part should be zero"
    denominator := (conjugate * complex) real.
    ^ComplexNumber
        real: (numerator real / denominator )
        imaginary: (numerator imaginary / denominator )! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 14:59'!
abs
   "Answer the absolute value of the receiver. "
    ^self modulus! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:00'!
amplitude
   "Answer the angle in radians for the trigonometric form of receiver"
    | pi pi2 n theta |
    pi := Float pi.
    pi2 := pi * 2.0.
    real = 0
        ifTrue: [
            imaginary >= 0
                ifTrue: [ theta := pi / 2.0.]
                ifFalse:[ theta := pi / 2.0 + pi.].
            ^theta ].
    theta := ( imaginary / real ) arcTan abs.
    theta  > pi2
      ifTrue:[ "adjust theta to be between 0 and 360 degrees"
        n := theta // pi2.
        theta := theta - ( pi2 * n ). ].
    imaginary > 0  "adjust angle based on signs of real,imaginary"
        ifTrue:[
            real < 0
                ifTrue: [ theta := pi - theta. ]. ]
        ifFalse:[
            real < 0
                ifTrue: [ theta := pi + theta ]
                ifFalse:[ theta := pi2 - theta ]. ].
    ^theta! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:04'!
ln
   "Answer the natural log of self"
    | r i theta |
    real = 0
      ifTrue:[
        imaginary > 0
            ifTrue:[ ^ComplexNumber
                        real: imaginary ln
                        imaginary: Float pi / 2.0 ]
            ifFalse:[ ^ComplexNumber
                        real: imaginary abs ln
                        imaginary: Float pi / -2.0  ].    ].
    real > 0
      ifTrue:[ theta := 0.0 ]
      ifFalse:[ " real < 0 "
        imaginary < 0
          ifTrue:[ theta := Float pi negated ]
          ifFalse:[ theta := Float pi ]. ].
    i := (imaginary / real) arcTan + theta.
    r := self modulus ln.
    ^ComplexNumber real: r imaginary: i! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:05'!
negated
   "Answer a new instance of self with real,imaginary numbers negated"
    ^ComplexNumber
        real: ( real negated)
        imaginary: (imaginary negated )! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:08'!
pseudoAbs
   "Answer the PSEUDO absolute value.
    Used for matrix computations.  Number supports same method"
    ^real abs + imaginary abs! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:09'!
raisedTo: aNumber
   "Answer the result receiver raised to the aNumber power.
    Uses De Moivre's theorem which employs the trigonometric form
    of a complexNumber."
    aNumber = 0
        ifTrue:[ ^1 ].
    ^self asComplexPolarForm raisedTo: aNumber! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:10'!
reciprocal
    "Answer the reciprocal of self."
    ^1 / self! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:11'!
sqrt
   "Answer one of the possible square roots of self"
    ^self raisedTo: 1/2! !

!ComplexNumber methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:12'!
squared
    "Answer the receiver squared."
    ^self * self! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:52'!
adaptToFloat: aFloat andSend: selector
	"If I am involved in arithmetic with aFloat, convert it to a ComplexNumber"

	^aFloat asComplexNumber perform: selector with: self! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:57'!
adaptToFraction: aFraction andSend: selector
	"If I am involved in arithmetic with aFloat, convert it to a ComplexNumber"

	^aFraction asComplexNumber perform: selector with: self! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:29'!
adaptToInteger: anInteger andSend: selector
	"If I am involved in arithmetic with an Integer, convert it to a ComplexNumber"

	^anInteger asComplexNumber perform: selector with: self! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:00'!
asComplexNumber
   "Answer self "
    ^self! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:01'!
asComplexPolarForm
    "Answer the trigonometric form of the receiver"
    ^ComplexPolarForm
        modulus: ( self modulus)
        amplitude: (self amplitude)! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:01'!
asFloat
   "Answer the real & imaginary parts asFloat
    makes Float operations *,+,-  changes valid"
    ^ComplexNumber
        real: real asFloat
        imaginary: imaginary asFloat! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:01'!
complexRoots: anInteger
   "Answer a Vector with the n-nth roots of self where
        n is anInteger > 0.      Uses De Moivre's theorem."
    ^self asComplexPolarForm complexRoots: anInteger! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:11'!
roundTo: aNumber
   "Answer the receiver with the real & imaginary parts
    rounded to the nearest multiple of aNumber"
    ^ComplexNumber
            real: (real roundTo: aNumber )
       imaginary: (imaginary roundTo: aNumber)! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:11'!
rounded
   "Answer an new instance with the real & imaginary parts
    rounded to the nearest integer."
    ^ComplexNumber
            real: real rounded
       imaginary: imaginary rounded! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:12'!
truncateTo: aNumber
   "Answer the receiver with the real & imaginary parts
    truncated (towards zero) to the nearest multiple of aNumber."
    ^ComplexNumber
            real: (real truncateTo: aNumber)
       imaginary: (imaginary truncateTo: aNumber)! !

!ComplexNumber methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:12'!
truncated
   "Answer an new instance with the real & imaginary parts truncated."
    ^ComplexNumber
            real: (real truncated)
       imaginary: (imaginary truncated)! !

!ComplexNumber methodsFor: 'private' stamp: 'jpwj 7/19/2001 15:10'!
real: realPart imaginary: imaginaryPart
   "Private - set the instance variables representing
    the real and imaginary parts of a complex number in algebraic form."
    real := realPart.
    imaginary := imaginaryPart! !

!ComplexNumber methodsFor: 'printing' stamp: 'jpwj 7/19/2001 15:05'!
printOn: aStream
   "Put the ASCII representation of ComplexNumber on aStream"
    real printOn: aStream.
    aStream nextPutAll: 'i:'.
    imaginary printOn: aStream! !

!ComplexNumber methodsFor: 'printing' stamp: 'jpwj 7/19/2001 15:06'!
printOn: aStream width: width decimal: places
   "Put the character represention of the real & imaginary parts
    of self on the stream "
    aStream nextPutAll: (self printWidth: width decimal: places)! !

!ComplexNumber methodsFor: 'printing' stamp: 'jpwj 7/19/2001 15:08'!
printRounded: numberFractionDigits
   "Answer a string with the ASCII character representation
    of the real & imaginary parts of the receiver
    rounded to numberFractionDigits decimal places."
    ^(real printRounded: numberFractionDigits ),'i:',(imaginary printRounded: numberFractionDigits)! !

!ComplexNumber methodsFor: 'printing' stamp: 'jpwj 7/19/2001 15:08'!
printWidth: width decimal: places
   "Answer a string with the ASCII character representation
    of the real & imaginary parts of the receiver
    right justified in width to the number of decimal places."
    ^(real printWidth: width decimal: places ),'i:',(imaginary printWidth: width decimal: places)! !

!ComplexNumber methodsFor: 'testing' stamp: 'jpwj 7/19/2001 15:04'!
isScalar
   "Answer true since a complex number is mathematically a scalar"
    ^true! !


!ComplexNumbers class methodsFor: 'instance creation' stamp: 'jpwj 7/19/2001 15:59'!
new
   "Answer an error since this is an invalid method"

    ^self error: 'Invalid message for Complex Numbers.'! !

!ComplexNumbers class methodsFor: 'instance creation' stamp: 'jpwj 7/19/2001 15:59'!
new: anArgument
   "Answer an error since this is an invalid method"
 
    ^self error: 'Invalid message for Complex Numbers.'! !


!ComplexNumber class methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:55'!
complexRootsOfUnity: anInteger
   "Answer a Vector containing the n nth roots of unity(1).
    Special case of instance method #complexRoots:"
    | answer theta twoPi half realInteger |
    answer := Vector new: anInteger.
    twoPi := 2.0 * Float pi.
    realInteger := anInteger asFloat.
    anInteger even
      ifTrue:[ "answer are symmetric except for sign"
        half := anInteger / 2.
        1 to: half do:[ :j |
          theta := twoPi * (j - 1) / realInteger.
          answer at: j put: (theta cos i: theta sin).
          answer at: j + half put: (theta cos i: theta sin) negated.]. ]
      ifFalse:[
        1 to: anInteger do:[ :j |
          theta := twoPi * (j - 1) / realInteger.
          answer at: j put: (theta cos i: theta sin). ]. ].
    ^answer! !

!ComplexNumber class methodsFor: 'instance creation' stamp: 'jpwj 7/19/2001 14:56'!
real: realPart imaginary: imaginaryPart
   "Answer a new instance of ComplexNumber.
     Neither part can be aKindOf: ComplexNumbers
    "
    ((realPart isKindOf: ComplexNumbers )
        or: [ imaginaryPart isKindOf: ComplexNumbers ] )
            ifTrue:[ ^self error: self class name,' invalid values'].
    ^ComplexNumber basicNew
        real: realPart imaginary: imaginaryPart! !


!ComplexPolarForm methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:16'!
amplitude
   "Answer amplitude of self (theta in the polar form)."
    ^amplitude! !

!ComplexPolarForm methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:19'!
imaginary
   "Answer the imaginary part of the polar form"
    ^( amplitude sin * modulus )
! !

!ComplexPolarForm methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:19'!
modulus
   "Answer modulus of self (r of the polar form)."
    ^modulus
! !

!ComplexPolarForm methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 15:20'!
real
   "Answer the real part of the polar form"
    ^( amplitude cos * modulus )! !

!ComplexPolarForm methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:14'!
* aNumber
   "Answer self asComplexNumber * aNumber"
    ^self asComplexNumber * aNumber

! !

!ComplexPolarForm methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:14'!
+ aNumber
   "Answer self asComplexNumber + aNumber"
    ^self asComplexNumber + aNumber
! !

!ComplexPolarForm methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:15'!
- aNumber
   "Answer self asComplexNumber - aNumber"
    ^self asComplexNumber - aNumber
! !

!ComplexPolarForm methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:15'!
/ aNumber
   "Answer self asComplexNumber / aNumber"
    ^self asComplexNumber / aNumber
! !

!ComplexPolarForm methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:17'!
complexRoots: anInteger
   "Answer a Vector with the n-nth roots of self where
    n is anInteger > 0.      Uses De Moivre's theorem."
    | n  r i m theta pi2 angle rootVector rint |
    ((anInteger isKindOf: Integer) and: [anInteger > 0 ])
      ifFalse:[
      ^self error: self class name,' nthRoot: requires a positive Integer.'].
    pi2 := Float pi * 2.0.
    theta := amplitude.
    m := modulus raisedTo: ( anInteger reciprocal ).
    rootVector := Vector new: anInteger.
    rint := anInteger asFloat.
    1 to: anInteger do: [ :k |
        angle := ( ((k - 1) * pi2) + theta ) / rint.
        angle > pi2
            ifTrue:[ n := angle // pi2. angle := angle - (pi2 * n). ].
        r := angle cos * m.
        i := angle sin * m.
        rootVector at: k put: ( r i: i). ].
    ^rootVector! !

!ComplexPolarForm methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:20'!
raisedTo: aNumber
   "Answer a ComplexNumber resulting from the receiver raised
    to the aNumber power. Uses De Moivre's theorem which employs
    the trigonometric form of a complexNumber."
    | m theta pi2  n |
    pi2 := Float pi * 2.0.
    theta := amplitude * aNumber.
    theta > pi2
      ifTrue:[ "adjust theta to be between 0 and 360 degrees"
        n := theta // pi2.
        theta := theta - ( pi2 * n ). ].
    m := modulus raisedTo: aNumber.
    ^ComplexNumber
        real: ( theta cos * m)
        imaginary: ( theta sin * m)! !

!ComplexPolarForm methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:16'!
asComplexNumber
   "Answer the algebraic form"
    ^ComplexNumber
        real: self real
        imaginary: self imaginary! !

!ComplexPolarForm methodsFor: 'printing' stamp: 'jpwj 7/19/2001 15:20'!
printOn: aStream
   "Put the ASCII representation of ComplexPolarForm on aStream"
    modulus printOn: aStream.
    aStream nextPutAll: 'cis:'.
    amplitude printOn: aStream.! !

!ComplexPolarForm methodsFor: 'private' stamp: 'jpwj 7/19/2001 15:19'!
modulus: r amplitude: theta
   "Private - set the instance variables, modulus and amplitude"
    modulus := r.
    amplitude := theta.
    ^self! !

!ComplexPolarForm methodsFor: 'unclassified' stamp: 'jpwj 7/19/2001 15:18'!
hash
   "Answer anInteger hash value for receiver"
    ^modulus hash + amplitude hash! !


!ComplexPolarForm class methodsFor: 'instance creation' stamp: 'jpwj 7/19/2001 15:13'!
modulus: r amplitude: theta
   "Answer a new instance of ComplexPolarForm.
    r(cos (theta) + i sin(theta)) where r=modulus & theta=amplitude
    theta is an angle in radians."
    ((r isKindOf: Number)
        and: [ (theta isKindOf: Number) ] )
            ifFalse:[ ^self error: self class name,' invalid values'].
    ^ComplexPolarForm basicNew
        modulus: r amplitude: theta! !


!Number methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 15:49'!
conjugate
   "Answer the conjugate of the receiver which is self for a Real Number
    Added for ComplexNumbers "

    ^self! !

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/19/2001 15:41'!
complexRoots: anInteger
   "Answer a Vector with all complex roots of the receiver
    Added for ComplexNumbers  jpwj 6/1/89"

    ^self asComplexPolarForm complexRoots: anInteger! !

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/21/2001 13:36'!
pythagorean: aNumber
   "Answer the hypotenuse of a right triangle where a=self
    and b=aNumber.  Method avoids underFlow or overflow.
    Based on subroutine pathag from EISPACK version 3.
    Added for ComplexNumbers  Added for MathPack jpwj 6/1/89"
    | a b p r t s u |
    a := self abs.  b := aNumber abs.
    p := a max: b.
    ( a = 0 or: [ b = 0 ] )
        ifTrue: [ ^p  ].
    p = 0
        ifTrue:[ ^0 ].
    r := ((a min: b ) / p ) raisedTo: 2.
    [ (t := 4.0 + r) = 4.0 ]
        whileFalse:[
            s := r / t. u := 2.0 * s + 1.0.  p := u * p.
            r := ( (s/u) raisedTo: 2 ) * r. ].
    ^p! !

!Number methodsFor: 'testing' stamp: 'jpwj 7/19/2001 15:48'!
isScalar
   "Answer true since any subclass of Number is mathematically a scalar."

    ^true! !

!Number methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:50'!
asComplexNumber
    "Answer the receiver as a complex number in algebraic form
    Added for ComplexNumbers  jpwj 6/1/89"

    ^ComplexNumber
        real: self
        imaginary: 0! !

!Number methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:44'!
asComplexPolarForm
    "Answer the receiver as a complex number in trigonometric form
    Added for ComplexNumbers.  jpwj 6/1/89"
    | r theta |
    self = 0
        ifTrue:[ ^self error: self class name,
            ' for asComplexPolarForm' ].
    r := self abs.
    self > 0
        ifTrue:[ theta := 0]
        ifFalse:[ theta := Float pi].
    ^ComplexPolarForm
        modulus: r
        amplitude: theta! !

!Number methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:40'!
cis: theta
   "Answer a complex number in polar form where self
    is the modulus and theta is the amplitude (angle in radians)
    self should be greater than 0. Added for ComplexNumbers"

    ^ComplexPolarForm
        modulus: self
        amplitude: theta! !

!Number methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:45'!
forceToZero: ifLessThanEpsilon
   "Answer zero if absolute value of the receiver
    is less than epsilon else answer self.   jpwj 6/1/89"

    self abs < ifLessThanEpsilon
      ifTrue:[ ^0  ]
      ifFalse:[ ^self ]! !

!Number methodsFor: 'converting' stamp: 'jpwj 7/19/2001 15:39'!
i: aNumber
   "Answer a complex number where self is the real part and
    imaginary part is aNumber. Added for ComplexNumbers "

    ^ComplexNumber
        real: self
        imaginary: aNumber! !

!Number methodsFor: 'as yet unclassified' stamp: 'jpwj 7/19/2001 15:46'!
imaginary
   "Answer the imaginary part of a real number
    Added for ComplexNumbers jpwj 6/1/89"

    ^0! !

!Number methodsFor: 'as yet unclassified' stamp: 'jpwj 7/19/2001 15:43'!
real
   "Answer real part (self) of the receiver.
    Added for ComplexNumbers  jpwj 6/1/89"

    ^self! !

