'From Squeak3.0 of 4 February 2001 [latest update: #3545] on 21 July 2001 at 1:34:43 pm'!
Matrices2Dimensional variableSubclass: #TriDiagonalMatrix
	instanceVariableNames: 'transformation eigenvectors eigenvalues '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MathPack-Matrices'!

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/19/2001 14:41'!
signOf: aNumber 
   "Answer self with the sign of aNumber" 

    aNumber < 0 
        ifTrue:[ ^self abs negated ]. 
    ^self abs! !


!SquareMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:31'!
copyLowerHalfFrom: aMatrix 
    "Copy the lower half of aMatrix including the diagonal 
    into the receiver." 
    ((aMatrix class = SquareMatrix) 
      or: [ aMatrix class = LowerTriangularMatrix ] ) 
      ifFalse: [ ^self error: self class name, 
        ' copyLowerHalfFrom: aMatrix is NOT N by N.' ]. 
    (aMatrix self rows =  self columns) 
      ifFalse: [ ^self error: self class name, 
        ' copyLowerHalfFrom: aMatrix has incorrect dimensions.' ]. 
    1 to: self rows do:[ :r | 
        1 to: r do:[ :c | 
            self row: r col: c put: (aMatrix row: r col: c). ].]! !

!SquareMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:24'!
copyLowerHalfToUpperHalf 
   "Copy the lower half (elements below diagonal) of self 
        to correct position in the upper half. 
    Should ONLY be used with a symmetric matrix." 
    self rows = 1 
        ifTrue:[ ^self ]. 
    2 to: self rows do:[ :r | 
        1 to: (r - 1) do:[ :c | 
            self row: c col: r put: (self row: r col: c). ].]! !

!SquareMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:25'!
eigenvalues 
        "Answer the value of eigenvalues." 
    ^eigenvalues! !

!SquareMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:25'!
eigenvectors 
        "Answer the value of eigenvectors." 
    ^eigenvectors! !

!SquareMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:31'!
transformation: aValue 
   "Private - Set the value of transformation matrix" 
    transformation := aValue! !

!SquareMatrix methodsFor: 'EigenSystem' stamp: 'jpwj 7/19/2001 14:23'!
jacobiEigenSystem
   "Determines the eigenvalues & eigenvectors for a real symmetric
    matrix by Jacobi transformations.  The values/vectors are stored
    in the instance variables of the matrix.
    Answers the numberOfRotations required to find
        the eigenvalues/vectors"
    | c  numberOfRotations d v b z tau g h s t threshold sum imajor
      dipabs diqabs theta |
    "should have a check to make sure matrix is symmetric"
    (self isSymmetric and:[ self isReal ])
      ifFalse:[
        ^self error: 'method valid ONLY for a Real Symmetric Matrix'].
    numberOfRotations := 0.
    v := Matrix identity: self rows. "initialize v to identity matrix"
    b := self diagonal. "initialize b & d to diagonal of self"
    d := self diagonal.
    z := (Vector new: self rows) atAllPut: 0.0.
    imajor := 1.
    [ imajor < 51 ] whileTrue:[ "outer loop structure"
    sum := 0.0.
    1 to: (self rows - 1) do:[ :ip |
        (ip + 1) to: self rows do:[ :iq |
            sum := (self row: ip col: iq) abs + sum ].].
    sum = 0.0
    ifTrue:[ imajor := 99 ] "solution found, terminate whileTrue:"
    ifFalse:[ "continue computation."
        imajor < 4
            ifTrue:[ threshold := (0.2 * sum) / (self rows * self rows) ]
            ifFalse:[ threshold := 0.0 ].
        1 to: (self rows - 1) do:[ :ip |
          (ip + 1) to: self rows do:[ :iq |
            g := (self row: ip col: iq) abs * 100.0.
            dipabs := (d at: ip) abs.
            diqabs := (d at: iq) abs.
            ( imajor > 4
                    and:[ dipabs + g = dipabs
                    and:[ diqabs + g = diqabs ] ] )
                ifTrue:[ self row: ip col: iq put: 0.0. ]
                ifFalse:[
            (self row: ip col: iq) abs > threshold
                ifTrue:[
                    h := (d at: iq) - (d at: ip).
                    (h abs + g = (h abs))
                        ifTrue:[
                            t := (self row: ip col: iq) / h ]
                        ifFalse:[
                            theta := (0.5 * h ) / (self row: ip col: iq).
                            t := theta abs + (( theta * theta + 1.0 ) sqrt).
                            t := 1 / t.
                            theta < 0
                                ifTrue:[ t := t negated ].   ].
                    c := (( t * t + 1.0) sqrt ) reciprocal.
                    s := t * c.
                    tau := s / ( 1.0 + c).
                    h := t * (self row: ip col: iq).
                    z at: ip put: ( (z at: ip) - h).
                    z at: iq put: ( (z at: iq) + h).
                    d at: ip put: ( (d at: ip) - h).
                    d at: iq put: ( (d at: iq) + h).
                    self row: ip col: iq put: 0.0.
                    1 to: (ip - 1) do:[ :j | "rotations 1<=j<p"
                        g := self row: j col: ip.
                        h := self row: j col: iq.
                        self row: j col: ip put: ( g - (s *(h+(g*tau)))).
                        self row: j col: iq put: ( h + (s *(g-(h*tau)))).].
                    (ip + 1) to: (iq - 1) do:[ :j | "rotations p<j<q"
                        g := self row: ip col: j.
                        h := self row: j col: iq.
                        self row: ip col: j put: ( g - (s *(h+(g*tau)))).
                        self row: j col: iq put: ( h + (s *(g-(h*tau)))).].
                    (iq + 1) to: self rows do:[ :j | "rotations q<j<=row"
                        g := self row: ip col: j.
                        h := self row: iq col: j.
                        self row: ip col: j put: ( g - (s *(h+(g*tau)))).
                        self row: iq col: j put: ( h + (s *(g-(h*tau)))).].
                    1 to: self rows do:[ :j |
                        g := v row: j col: ip.
                        h := v row: j col: iq.
                        v row: j col: ip put: ( g - (s *(h+(g*tau)))).
                        v row: j col: iq put: ( h + (s *(g-(h*tau)))).].
                     numberOfRotations := numberOfRotations + 1.
                        ]."ifTrue self > threshold"
                    ]."ifFalse imajor > 4 etc"
            ].]."end ip,iq loops"
        1 to: self rows do:[ :j |
          b at: j put:( (b at: j) + (z at: j)). ].
        d := b copy.
        z atAllPut: 0.0.
        imajor := imajor + 1.]. "end ifFalse sum = 0.0"
    ]. "end outer loop structure"
    imajor = 99
      ifFalse:[ ^self error: self class name,
        ' jacobiTransformation method failure.' ].
    "restore upper half of self which was destroyed during computation"
    self copyLowerHalfToUpperHalf.
    eigenvalues := d.
    eigenvectors := v.
    self sortEigenSystem: 'vectors'.
    ^numberOfRotations! !

!SquareMatrix methodsFor: 'private' stamp: 'jpwj 7/19/2001 14:27'!
eigenvalues: aVector
   "Private - Store the array of eigenvalues."
    eigenvalues := aVector! !

!SquareMatrix methodsFor: 'private' stamp: 'jpwj 7/19/2001 14:27'!
eigenvectors: aMatrix
   "Private - Set the value of matrix of eigenvectors."
    eigenvectors := aMatrix! !

!SquareMatrix methodsFor: 'unclassified' stamp: 'jpwj 7/19/2001 14:30'!
jacobiTransformation
   "Determines the eigenvalues & eigenvectors for a real symmetric
matrix.  The values/vectors are stored in the instance variables
of the matrix.  Method answers the numberOfRotations required to find
the eigenvalues/vectors"
    | c  numberOfRotations d v b z tau g h s t threshold sum imajor
      dipabs diqabs theta |
    "should have a check to make sure matrix is symmetric"
    numberOfRotations := 0.
    v := Matrix identity: self rows. "initialize v to identity matrix"
    b := self diagonal. "initialize b & d to diagonal of self"
    d := self diagonal.
    z := (Vector new: self rows) atAllPut: 0.0.
    imajor := 1.
    [ imajor < 51 ] whileTrue:[ "outer loop structure"
    sum := 0.0.
    1 to: (self rows - 1) do:[ :ip |
        (ip + 1) to: self rows do:[ :iq |
            sum := (self row: ip col: iq) abs + sum ].].
    sum = 0.0
    ifTrue:[ imajor := 99 ] "solution found, terminate whileTrue:"
    ifFalse:[ "continue computation."
        imajor < 4
            ifTrue:[ threshold := (0.2 * sum) / (self rows * self rows) ]
            ifFalse:[ threshold := 0.0 ].
        1 to: (self rows - 1) do:[ :ip |
          (ip + 1) to: self rows do:[ :iq |
            g := (self row: ip col: iq) abs * 100.0.
            dipabs := (d at: ip) abs.
            diqabs := (d at: iq) abs.
            ( imajor > 4
                    and:[ dipabs + g = dipabs
                    and:[ diqabs + g = diqabs ] ] )
                ifTrue:[ self row: ip col: iq put: 0.0. ]
                ifFalse:[
            (self row: ip col: iq) abs > threshold
                ifTrue:[
                    h := (d at: iq) - (d at: ip).
                    (h abs + g = (h abs))
                        ifTrue:[
                            t := (self row: ip col: iq) / h ]
                        ifFalse:[
                            theta := (0.5 * h ) / (self row: ip col: iq).
                            t := theta abs + (( theta * theta + 1.0 ) sqrt).
                            t := 1 / t.
                            theta < 0
                                ifTrue:[ t := t negated ].   ].
                    c := (( t * t + 1.0) sqrt ) reciprocal.
                    s := t * c.
                    tau := s / ( 1.0 + c).
                    h := t * (self row: ip col: iq).
                    z at: ip put: ( (z at: ip) - h).
                    z at: iq put: ( (z at: iq) + h).
                    d at: ip put: ( (d at: ip) - h).
                    d at: iq put: ( (d at: iq) + h).
                    self row: ip col: iq put: 0.0.
                    1 to: (ip - 1) do:[ :j | "rotations 1<=j<p"
                        g := self row: j col: ip.
                        h := self row: j col: iq.
                        self row: j col: ip put: ( g - (s *(h+(g*tau)))).
                        self row: j col: iq put: ( h + (s *(g-(h*tau)))).].
                    (ip + 1) to: (iq - 1) do:[ :j | "rotations p<j<q"
                        g := self row: ip col: j.
                        h := self row: j col: iq.
                        self row: ip col: j put: ( g - (s *(h+(g*tau)))).
                        self row: j col: iq put: ( h + (s *(g-(h*tau)))).].
                    (iq + 1) to: self rows do:[ :j | "rotations q<j<=row"
                        g := self row: ip col: j.
                        h := self row: iq col: j.
                        self row: ip col: j put: ( g - (s *(h+(g*tau)))).
                        self row: iq col: j put: ( h + (s *(g-(h*tau)))).].
                    1 to: self rows do:[ :j |
                        g := v row: j col: ip.
                        h := v row: j col: iq.
                        v row: j col: ip put: ( g - (s *(h+(g*tau)))).
                        v row: j col: iq put: ( h + (s *(g-(h*tau)))).].
                     numberOfRotations := numberOfRotations + 1.
                        ]."ifTrue self > threshold"
                    ]."ifFalse imajor > 4 etc"
            ].]."end ip,iq loops"
        1 to: self rows do:[ :j |
          b at: j put:( (b at: j) + (z at: j)). ].
        d := b copy.
        z atAllPut: 0.0.
        imajor := imajor + 1.]. "end ifFalse sum = 0.0"
    ]. "end outer loop structure"
    imajor = 99
      ifFalse:[ ^self error: self class name,
        ' jacobiTransformation method failure.' ].
    "restore upper half of self which was destroyed during computation"
    self copyLowerHalfToUpperHalf.
    eigenvalues := d.
    eigenvectors := v.
    self sortEigenSystem: 'vectors'.
    ^numberOfRotations! !

!SquareMatrix methodsFor: 'unclassified' stamp: 'jpwj 7/19/2001 14:26'!
realSymmetricEigenSystem: option 
   "Answer self with results stored in the instance variables. 
    option := 'values' for computing ONLY eigenvalues. 
    option := 'vectors' for computing both eigenvalues and eigenvectors." 
    | opt triDiagonal | 
    (self isSymmetric and:[ self isReal ]) 
      ifFalse:[ 
        ^self error: 'method valid ONLY for a Real Symmetric Matrix']. 
    opt := option asLowercase. 
    opt = 'values' 
        ifTrue: [ opt := nil] 
        ifFalse:[ 
            opt = 'vectors' 
                ifFalse:[ self error: self class name,' invalid option'].]. 
    triDiagonal := self reduceToTriDiagonal: opt. 
    triDiagonal computeRealSymmetricEigenSystem: opt. 
    triDiagonal sortEigenSystem: opt. 
    eigenvalues := triDiagonal eigenvalues. 
    opt isNil 
        ifFalse:[ eigenvectors := triDiagonal eigenvectors]! !

!SquareMatrix methodsFor: 'unclassified' stamp: 'jpwj 7/19/2001 14:29'!
reduceToTriDiagonal: option 
   "Answer the triDiagonalMatrix resulting from orthogonal similiarity 
transformations of a REAL SYMMETRIC matrix.  Similiar to TRED2 (old version) 
in EISPACK. 
    option isNil compute information for eigenvalues ONLY 
           else compute informaton for both eigenvalues/vectors." 
    | d e z l  h scale f g jp1 hh triDiagonal | 
        "Special Case" 
    self rows = 1 
        ifTrue:[ ^self error: self class name, 
            ' method invalid for 1x1' ]. 
    z := self copy. "working copy of self" 
    e := (Vector new: self rows) atAllPut: 0.0. 
    d := (Vector new: self rows) atAllPut: 0.0. 
    "**MAIN LOOP**" 
    self rows to: 2 by: -1 do:[ :i | 
        l := i - 1.   h := 0.0.    scale := 0.0. 
    l < 2 
    ifTrue:[ 
        e at: i put: (z row: i col: l). ] 
    ifFalse:[ "(l < 2)" 
        1 to: l do:[ :k | 
            scale := (z row: i col: k) abs + scale. ]. 
     scale = 0 
        ifTrue:[ 
            e at: i put: (z row: i col: l). ] 
        ifFalse:[ 
            1 to: l do:[ :k |  "do 150" 
                z row: i col: k  put: (z row: i col: k) / scale. 
                h  := h + ((z row: i col: k) * (z row: i col: k)). ]. 
            f := z row: i col: l. 
            g := ((h sqrt ) signOf: f ) negated. 
            e at: i put: scale * g. 
            h := h - (f * g). 
            z row: i col: l put: (f - g). 
            f := 0.0. 
            1 to: l do:[ :j | "do 240" 
                z row: j col: i put: ((z row: i col: j) / h). 
                g := 0.0. 
                1 to: j do:[ :k |  "do 180"  "------- form a*u ------" 
                    g := g + ( (z row: j col: k) * (z row: i col: k) ).]. 
                jp1 := j + 1. 
                (l < jp1) 
                  ifFalse:[ 
                    jp1 to: l do:[ :k | "do 200" 
                        g := g + ((z row: k col: j) * (z row: i col: k)).].]. 
                e at: j put: g / h. "--------- form p --------" 
                f := f + ( (e at: j) * (z row: i col: j) ). ]. "end do 240" 
            hh := f / ( h + h). 
                "--------- form reduced z -------" 
            1 to: l do:[ :j | "do 260" 
                f := z row: i col: j. 
                g := (e at: j) - ( hh * f). 
                e at: j put: g. 
                1 to: j do:[ :k | 
                    z row: j col: k put: ( (z row: j col: k) - 
                        ( f * (e at: k)) - ( g * (z row: i col: k)) ).]. ]. 
                ].  "end ifFalse: scale = 0" 
            ]. "end ifFalse(l < 2)" 
        d at: i put: h. 
    ]."**End Main Loop**" 
"----------- Accumulation of transformation matrices ------------" 
    d at: 1 put: 0.0. 
    e at: 1 put: 0.0. 
    1 to: self rows do:[ :i | 
    option isNil ifFalse:[ 
        l := i - 1. 
        (d at: i) = 0 
            ifFalse:[ 
                1 to: l do:[ :j | "do 360" 
                    g := 0.0. 
                    1 to: l do:[ :k | "do 340" 
                        g := g + ( (z row: i col: k) * (z row: k col: j) ).]. 
                    1 to: l do:[ :k | "do 360" 
                        z row: k col: j put: ( (z row: k col: j) 
                            - (g * (z row: k col: i)) ).].]. 
                 ]. "end ifFalse(h=0)" ]. "end option isNil" 
 
        d at: i put: ( z row: i col: i). 
 
        option isNil ifFalse:[ 
        z row: i col: i put: 1.0. 
        (l < 1 ) 
            ifFalse:[ 
                1 to: l do:[ :j | "do 400" 
                    z row: j col: i put: 0.0. 
                    z row: i col: j put: 0.0 ].].]. "end option isNil" 
        ]. "end accumlation of transformation matrices" 
    "Create a tridiagonal matrix, put in d and e" 
    triDiagonal := TriDiagonalMatrix new: self rows @ self rows. 
    triDiagonal diagonal: d. 
    triDiagonal subdiagonal: e. 
        "Save the transformation matrix?" 
    option isNil 
        ifTrue:[ z := nil ]. 
    triDiagonal transformation: z. 
    ^triDiagonal! !

!SquareMatrix methodsFor: 'unclassified' stamp: 'jpwj 7/19/2001 14:24'!
sortEigenSystem: option 
   "Sort the eigenvalues and eigenvectors into descending order 
    by the eigenvalues.  if option isNil then sort only eigenvalues 
    else option (isNotNil) sort eigenvalues & eigenvectors " 
    | k p | 
    1 to: self rows - 1 do:[ :i | 
        k := i. 
        p := eigenvalues at: i. 
        (i + 1) to: self rows do:[ :j | "find largest value" 
             (eigenvalues at: j) >= p 
                ifTrue:[ k := j.  p := eigenvalues at: j. ]. ]. 
        ( k = i ) 
            ifFalse:[ 
                eigenvalues at: k put: (eigenvalues at: i). 
                eigenvalues at: i put: p. 
                option isNil 
                  ifFalse:[ eigenvectors column: k exchange: i. ]. 
            ]. "end ifFalse: k = i" 
    ]! !


!TriDiagonalMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:39'!
eigenvalues 
        "Answer the value of eigenvalues." 
    ^eigenvalues! !

!TriDiagonalMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:38'!
eigenvectors 
        "Answer the value of eigenvectors." 
    ^eigenvectors! !

!TriDiagonalMatrix methodsFor: 'accessing' stamp: 'jpwj 7/19/2001 14:38'!
transformation 
        "Answer the value of transformation." 
    ^transformation! !

!TriDiagonalMatrix methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 14:38'!
computeRealSymmetricEigenSystem: option 
   "Compute eigenValues and eigenVectors from the transformation 
    matrix and the triDiagonal matrix generated with reduceToTriDiagonal 
    results are stored in instance variables eigenvalues, eigenvectors 
    method answers the errorFlag 
    errorFlag = 0 - no error 
    errorFlag > 0  the number of the eigenvalue that the computation 
        was trying to find when it failed. 
        modeled after EISPACK routine imtql2" 
    | d e z errorFlag  m b c f g p r s tst1 tst2 
      iterationCount k | 
    self rows = 1 
        ifTrue:[ ^self error: self class name, 
    'computeEigenSystem invalid on 1x1']. 
    d := self diagonal. 
    e := self subdiagonalN. 
    z := transformation copy. 
    errorFlag := 0. 
    2 to: self rows do:[ :i | 
        e at: (i - 1) put: (e at: i).]. 
    e at: self rows put: 0.0. 
        "Top of main do loop" 
    1 to: self rows do:[ :l | 
        iterationCount := 0. 
            "locate inital index(m) of small sub-diagonal element" 
        k := l. 
        m := self rows. "set default index" 
        [ k < self rows ] whileTrue:[ 
            tst1 := (d at: k) abs + ((d at: ( k + 1)) abs). 
            tst2 := (e at: k) abs + tst1. 
            ( tst1  = tst2 ) 
                ifTrue:[ 
                    m := k. "set index" 
                    k := self rows. "force termination of do while"]. 
            k := k + 1. ]. 
        [ m ~= l ] whileTrue:[ "iterate to find solution" 
        iterationCount :=  iterationCount + 1. 
        iterationCount > 30 
            ifTrue:[ "Failure to converge, save computations then exit" 
                eigenvalues := d. 
                eigenvectors := z. 
                errorFlag := l. 
                ^errorFlag ]. 
            "computation of eigenvalues" 
        p := d at: l. 
        g := ( (d at:(l + 1)) - p ) / ( 2.0 * (e at: l)). 
        r :=  g pythagorean: 1.0. 
        g := (( d at: m ) - p ) + ( (e at: l) / ( g + (r signOf: g))). 
        s := 1.0.  c := 1.0.  p := 0.0. 
        ( m - 1 ) to: l by: -1 do:[ :i | 
            f := s * (e at: i). 
            b := c * (e at: i). 
            ( f abs >= (g abs) ) 
                ifTrue:[ 
                    c := g / f. 
                    r := c pythagorean: 1.0. 
                    e at: (i + 1) put: ( f * r). 
                    s := 1.0 / r. 
                    c := c * s. ] 
                ifFalse:[ 
                    s := f / g. 
                    r := s pythagorean: 1.0. 
                    e at: (i + 1) put: ( g * r). 
                    c := 1.0 / r. 
                    s := c * s. ]. 
            g := (d at: (i + 1)) - p. 
            r := (((d at: i) - g) * s) + ( 2.0 * c * b). 
            p := s * r. 
            d at: (i + 1 ) put: (g + p). 
            g := (c * r ) - b. 
        option isNil ifFalse:[ 
            1 to: self rows do:[ :kk | "form eigenvectors" 
                f := z row: kk col: (i + 1). 
                z row: kk col: (i + 1) put: 
                    ( (s * (z row: kk col: i)) + ( c * f) ). 
                z row: k col: i put: 
                    ( (c * (z row: kk col: i)) - ( s * f) ). ]. 
              ]. "end ( m - 1 ) to: l by: -1 do:[ :i | " 
                    ]."end option isNil" 
 
            d at: l put: ( (d at: l) - p). 
            e at: l put: g. 
            e at: m put: 0.0. 
            "locate next index(m) of small sub-diagonal element" 
        k := l. 
        m := self rows. "set default index" 
        [ k < self rows ] whileTrue:[ 
            tst1 := (d at: k) abs + ((d at: ( k + 1)) abs). 
            tst2 := (e at: k) abs + tst1. 
            ( tst1  = tst2 ) 
                ifTrue:[ 
                    m := k. "set index" 
                    k := self rows. "force termination of do while"]. 
            k := k + 1. ]. 
        ]. "end m = l whileFalse" 
    ]. "end main do loop" 
    "save eigenvalues/vectors" 
    eigenvalues := d. 
    eigenvectors := z. 
        "Done answer the errorFlag" 
    ^errorFlag! !

!TriDiagonalMatrix methodsFor: 'private' stamp: 'jpwj 7/19/2001 14:34'!
eigenvalues: aVector
   "Private - Store the array of eigenvalues."
    eigenvalues := aVector! !

!TriDiagonalMatrix methodsFor: 'private' stamp: 'jpwj 7/19/2001 14:40'!
eigenvectors: aMatrix
   "Private - Set the value of matrix of eigenvectors."
    eigenvectors := aMatrix! !

!TriDiagonalMatrix methodsFor: 'private' stamp: 'jpwj 7/19/2001 14:34'!
transformation: aValue 
   "Private - Set the value of transformation matrix" 
    transformation := aValue! !

!TriDiagonalMatrix methodsFor: 'unclassified' stamp: 'jpwj 7/19/2001 14:39'!
sortEigenSystem: option 
   "Sort the eigenvalues and eigenvectors into descending order 
    by the eigenvalues" 
    | k p | 
    1 to: self rows - 1 do:[ :i | 
        k := i. 
        p := eigenvalues at: i. 
        (i + 1) to: self rows do:[ :j | "find largest value" 
             (eigenvalues at: j) >= p 
                ifTrue:[ k := j.  p := eigenvalues at: j. ]. ]. 
        ( k = i ) 
            ifFalse:[ 
                eigenvalues at: k put: (eigenvalues at: i). 
                eigenvalues at: i put: p. 
                option isNil 
                  ifFalse:[ eigenvectors column: k exchange: i. ]. 
            ]. "end ifFalse: k = i" 
    ]! !

Matrices2Dimensional variableSubclass: #TriDiagonalMatrix
	instanceVariableNames: 'eigenvalues eigenvectors transformation '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MathPack-Matrices'!
