'From Squeak3.0 of 4 February 2001 [latest update: #3545] on 21 July 2001 at 1:32:36 pm'!

!Vector commentStamp: 'jpwj 7/20/2001 08:14' prior: 0!
Vector is a subclass of Array for handling 'Arrays' of numbers
The methods added are for numerical computations.
Using the new methods means that the values stored in the Vector MUST
be of class Number(ComplexNumber) or the UndefinedObject (nil).
The UndefinedObject nil is allowed to accomodated the real world
situation of missing values.  nil's are ignored in most computational methods.
  Notes on nil:
Use of lowercase nil creates the desired UndefinedObject in
the following cases:
   - Changing the value of an vector element in an
        Inspector window to nil.
    - at: anInteger put: nil.
    - anVector := #( 1 nil 2 nil ) asVector convertNilSymbols.
Usage of nil that will NOT produce an UndefinedObject.
- anVector := #( 1 nil 2 nil ) asVector.  results in the symbol, nil.
  method convertNilSymbols converts nil symbols to UndefinedObject.

*, +, -, / methods will accept a scalar or another Vector of the
same size as the receiver. The result will be a new vector.
These operations are done element wise when the object is an Vector.
e.g. c := a + b.  c will be an Vector with same size as a.
the contents of c are computed: c at: i put: ((a at: i) + (b at: i)).
when b is an Vector else  c at: i put: ((a at: i) + b ) for each
non-nil element in a.   When a & b are Vectors then nil is stored
in c at: i if either value (a at: i) or (b at: i) is nil.

 !


!Collection methodsFor: 'converting' stamp: 'jpwj 7/19/2001 16:25'!
asVector
   "Answer a Vector containing all the elements of the receiver.
        Added for MathPack jpwj 6/1/89"
    | answer index |
    answer := Vector new: self size.
    index := 1.
    self do: [ :element |
        answer at: index put: element.
        index := index + 1].
    ^answer! !


!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:23'!
beta: w
   "Answer the value of the beta function.
    Passed message on to Float.     MathPack jpwj 6/1/89"

    ^self asFloat beta: w! !

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:23'!
betaIncomplete: aValue b: bValue
    "Answer the value of the incomplete beta function.
    Passed message on to Float.    MathPack jpwj 6/1/89"

    ^self asFloat betaIncomplete: aValue b: bValue! !

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/21/2001 13:28'!
gammaIncomplete: x
   "Answer the value for gammaIncomplete(a,x) function
        where a is self
    Pass message on to Float.  Added MathPack jpwj 6/1/89"

    ^self asFloat gammaIncomplete: x! !

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:19'!
gammaIncompleteComplement: x
   "Answer the value for complement of gammaIncomplete(a,x) function
        where a is self
    Pass message on to Float.    Added for MathPack jpwj 6/1/89"

    ^self asFloat gammaIncompleteComplement: x! !

!Number methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:24'!
gammaLn
    "Answer the ln of the gamma function on self.
    Passed message on to Float.    MathPack jpwj 6/1/89"

    ^self asFloat gammaLn! !

!Number methodsFor: 'printing' stamp: 'jpwj 7/20/2001 08:20'!
printOn: aStream width: width decimal: places
   "Put the character represention of self on the stream
    in the width given with the number of decimal places.
    If it will not fit put asterisks(***)  Added for MathPack jpwj 6/1/89"

    aStream nextPutAll: (self printWidth: width decimal: places)! !

!Number methodsFor: 'printing' stamp: 'jpwj 7/20/2001 12:01'!
printRounded: numberFractionDigits
   "Answer a string with the ASCII character representation
    of the receiver rounded to numberFractionDigits decimal places.
Extented to allow more than eight decimal places.
Fixed to V286 R1.1 method, last two rounder values had one too many zeros
jpwj 6/1/89
This method still needs some more thought - Squeak may have better ways."
    | rounder |
    numberFractionDigits > 7
        ifFalse: [ rounder := 1/(#(1 10 100 1000 10000
                        100000 1000000 10000000)
                    at: numberFractionDigits + 1).]
        ifTrue:[ rounder := 1 / (10 raisedToInteger: numberFractionDigits + 1)].
    ^(self roundTo: rounder asFloat) printString
        "printFraction: numberFractionDigits"! !

!Number methodsFor: 'printing' stamp: 'jpwj 7/20/2001 08:18'!
printWidth: width decimal: places
    "Convert Number to a string of given width with decimal places.
     Result is right justified.  String filled with asterisks
     if conversion will not fit in the with given. MathPack jpwj 6/1/89"
    | string s1 index |
    string := (String new: width) atAllPut: $ .
    places > 0
        ifTrue: [ s1 := (self asFloat) printRounded: places.]
        ifFalse:[ s1 := (self rounded) printString].
    s1 size > width
      ifTrue: [ string atAllPut: $* ]
      ifFalse:[ index := width.
        s1 size to: 1 by: 0 - 1 do: [ :i |
          string at: index put: (s1 at: i).
          index := index - 1. ].].
    ^string! !

!Number methodsFor: 'as yet unclassified' stamp: 'jpwj 7/20/2001 08:17'!
errorFunction
   "Answers the special case of the gammaIncomplete function
    Pass message on to Float.   Added for MathPack jpwj 6/1/89"

    ^self asFloat errorFunction! !

!Number methodsFor: 'as yet unclassified' stamp: 'jpwj 7/20/2001 08:29'!
errorFunctionComplement
   "Answers complementary error function
    Pass message on to Float.    MathPack jpwj 6/1/89"

    ^self asFloat errorFunctionComplement! !


!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:38'!
beta: w
   "Answer the value of the beta function defined as
     beta(z,w) = gamma(z)gamma(w) / gamma(z + w) where z is self
    		MathPack jpwj 6/1/89"

    ^( self gammaLn + w gammaLn - ((self + w) gammaLn )) exp! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:34'!
betaIncomplete: aValue b: bValue
    "Answer the value of the incomplete beta function defined
        Ix(a,b).  incomplete beta at x for values (a,b)
            where x is self    MathPack jpwj 6/1/89"
    | a b temp answer |
    a := aValue asFloat. b := bValue asFloat.
    (( self < 0.0 ) or: [ self > 1.0 ] )
        ifTrue: [ self error: self class name,
            ' Invalid value for betaIncomplete:b: '].
    (( self = 0.0 ) or: [ self = 1.0 ] )
        ifTrue: [ temp := 0.0 ]
        ifFalse: [ temp := ((a+b) gammaLn ) - (a gammaLn)
            - (b gammaLn) + ( a * (self ln)) + (b * ((1.0 - self) ln) ).
            temp := temp exp. ].
    self < ((a + 1.0) / (a+b+2.0))
      ifTrue: [
        answer := temp * (self betaContinuedFraction: a b: b ) / a ]
      ifFalse: [
        answer := (1.0 - self) betaContinuedFraction: b b: a.
        answer := 1.0 - (temp * answer / b). ].
    ^answer! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:31'!
errorFunction
   "Answers the special case of the gammaIncomplete function
    errorFunction(x)  = gammaIncomplete(0.5,x*x).
        where x is self.    MathPack jpwj 6/1/89"
    | answer |
    answer := 0.5 gammaIncomplete: (self * self).
    self < 0
      ifTrue:[ ^answer negated ]
      ifFalse:[ ^answer ]! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:37'!
errorFunctionComplement
   "Answer the complementary error function associated with self
    errorFunctionComplement(x) = 1 - errorFunction(x)
        where x is self.    MathPack jpwj 6/1/89"
    self < 0
        ifTrue: [ ^(1.0 + (0.5 gammaIncomplete: ( self * self)) ) ]
        ifFalse:[ ^(0.5 gammaIncompleteComplement: ( self * self)) ].! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:35'!
fRatioSignificance: numberDegreesOfFreedom1 f2: numberDegreesOfFreedom2
    "Answer the significance of the f-ratio (self) with the two associated
    number of degrees of freedom.  The f-ratio is the larger variance
    divided by smaller variance of two data vectors. numberDegreesOfFreedom1
    is associated with larger variance and numberDegreesOfFreedom2 with the
    smaller variance.  Degrees of freedom are the number of elements (nonNil)
    in the respective data vector less one.
        Answers with small values indicate significantly different variances.
    (1.0 - answer) is the F-Distribution Probability Function value.
    (Used for computing F-Test for Significantly Different Variances.)    MathPack jpwj 6/1/89"
    | f1 f2 answer x1 x2  |
    f1 := numberDegreesOfFreedom1 asFloat.
    f2 := numberDegreesOfFreedom2 asFloat.
    x1 := f2 / ( f2 + (f1 * self)).
    x2 := f1 / ( f1 + (f2 / self)).
    answer :=  ( x1 betaIncomplete: (0.5 * f2) b: (0.5 * f1))
              +( 1.0 - (x2 betaIncomplete: (0.5 * f1) b: (0.5 * f2))).
    ^answer * 0.5! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:36'!
gammaIncomplete: x
   "Answer the value for gammaIncomplete(a,x) function
        where a is self    MathPack jpwj 6/1/89"
    ( x < 0 or:[ self <= 0 ])
      ifTrue:[^self error: self class name,
            ' gammaIncomplete: invalid values'].
    x < (self + 1)
        ifTrue: [ ^(self gammaIncompleteBySeries: x) ]
        ifFalse:[ ^( 1.0 - (self gammaIncompleteByContinuedFraction: x) )]! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:34'!
gammaIncompleteByContinuedFraction: x
   "Private - Used by gammaIncomplete:
    Answer the imcomplete gamma function by continued fraction evaluation
    		MathPack jpwj 6/1/89"
    | g  a0 a1 b0 b1 fac an ana anf gold |
    gold := 0.0.  a0 := 1.0. a1 := x. b0 := 0.0. b1 := 1.0. fac := 1.0.
    an := 1.0.
    [ an <= 100 ] whileTrue:[
        ana := an - self.
        a0  := (a1 + (a0 * ana)) * fac.
        b0  := (b1 + (b0 * ana)) * fac.
        anf := an * fac.
        a1  := x * a0 + (anf * a1).
        b1  := x * b0 + (anf * b1).
        a1 ~= 0
          ifTrue:[
            fac := a1 reciprocal.
            g := b1 * fac.
            ((g - gold) abs / g ) < 3.0e-7
                ifTrue:[
                    gold := (x negated + ( self * x ln) - self gammaLn) exp * g.
                    ^gold ].
            gold := g ].
        an := an + 1.0. ].
    "failed to converge either self is too large OR
            interation count reached 100 "
    ^self error: self class name,
        ' gammaIncompleteByContinuedFraction: failed to converge'! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:32'!
gammaIncompleteBySeries: x
   "Private - Used by gammaIncomplete:
    Answer the imcomplete gamma function by its series expansion
        MathPack jpwj 6/1/89"
    | aPlusOne sum del count |
    x = 0
      ifTrue:[ ^0 ].
    x < 0
      ifTrue:[^self error: self class name,
            ' gammaIncompleteBySeries: invalid value'].
    aPlusOne := self.
    sum := self reciprocal.
    del := sum.
    count := 0.
    [ count < 100 ] whileTrue:[
        aPlusOne := aPlusOne + 1.0.
        del := del * x / aPlusOne.
        sum := sum + del.
        del abs < ( sum abs * 3.0e-7 )
            ifTrue:[^((x negated + (self * x ln) - self gammaLn) exp * sum)].
        count := count + 1. ].
        "failed to converge either self is too large OR
            interation count reached 100 "
    ^self error: self class name,
        ' gammaIncompleteBySeries: failure to converge'! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:33'!
gammaIncompleteComplement: x
   "Answer the value of gammaIncompleteComplement(a,x) function
    defined as gammaIncompleteComplement(a,x) = 1 - gammaIncomplete(a,x).
        where a is self.    MathPack jpwj 6/1/89"
    ( x < 0 or:[ self <= 0 ])
      ifTrue:[^self error: self class name,
            ' gammaIncompleteComplement: invalid values'].
    x < (self + 1)
        ifTrue: [ ^(1.0 - (self gammaIncompleteBySeries: x)) ]
        ifFalse:[ ^(self gammaIncompleteByContinuedFraction: x)]! !

!Float methodsFor: 'mathematical functions' stamp: 'jpwj 7/20/2001 08:40'!
gammaLn
    "Answer the ln of the gamma function on self.    MathPack jpwj 6/1/89"

    self > 0
      ifFalse:[^self error: self class name,' < 0, invalid for gammaLn'].
    ^self gammaLn1! !

!Float methodsFor: 'private' stamp: 'jpwj 7/20/2001 08:41'!
betaContinuedFraction: aValue b: bValue
    "Private - Continued fraction approximation for
     betaIncomplete function    MathPack jpwj 6/1/89"

    | a b am bm az qab qap qam bz em tem d aold ap app bp bpp |
    a := aValue asFloat. b := bValue asFloat.
    am := bm := az := 1.0. qab := a + b.  qap := a + 1.0.
    qam := a - 1.0.  bz := 1.0 - ( (qab*self) / qap).
    em := 0 asFloat.
    [ em < 100.0 ] whileTrue: [
        em := em + 1.0.
        tem := em + em.
        d := (em * (b - em) * self) / ((qam + tem) * (a + tem)).
        ap := az + (d*am).
        bp := bz + (d*bm).
        d := ((a+em) negated * (qab+em) * self) / ((a+tem)*(qap+tem)).
        app := ap + (d*az).
        bpp := bp + (d*bz).
        aold := az.
        am := ap / bpp.  bm := bp / bpp.  az := app / bpp.
        bz := 1 asFloat.
        ( (az - aold) abs ) < ( 3.0e-7 * (az abs) )
            ifTrue: [ ^az ]. ].
    ^self error: self class name,'betaContinuedFraction failed to converge'

! !

!Float methodsFor: 'private' stamp: 'jpwj 7/20/2001 08:39'!
gammaLn1
   "Private - Used by gammaLn.  Answers the value of gammaLn(self)
    function for self > 1 through series expansion.   MathPack jpwj 6/1/89 "
    | coefficents x tmp seriesSum |
    coefficents := #( 76.18009173e0  -86.50532033e0 24.01409822e0
            -1.231739516e0 0.120858003e-2  -0.536382e-5 ).
    x := self - 1.0.  tmp := x + 5.5.
    x = 0.0
      ifTrue:[ ^0.0 ].
    tmp := ((x + 0.5) * (tmp ln)) - tmp.
    seriesSum := 1.0.
    1 to: 6 do: [ :i |
        x := x + 1.0.
        seriesSum := seriesSum + ( (coefficents at: i) / x ) ].
    ^tmp + ( (2.50662827465e0 * seriesSum) ln).! !


!Integer methodsFor: 'mathematical functions' stamp: 'jpwj 7/21/2001 13:29'!
binomialCoefficient: k
   "Answer the binomial coefficient | n |
                                    | k |
        where 0 <= k <= n   and n is self    Added for MathPack jpwj 6/1/89"
    k < 0
      ifTrue:[ ^self error: self class name,
        ' binomialCoefficient: requires positive values'].
    self < k
      ifTrue:[ ^self error: self class name,
        ' binomialCoefficient: k > n '].
    ^( self factorial / (k factorial * (self - k) factorial ))! !

!Integer methodsFor: 'mathematical functions' stamp: 'jpwj 7/19/2001 16:16'!
gammaLn
   "Answer the ln of the gamma function for an Integer value
    i.e. n factorial = gamma(n + 1)  for MathPack jpwj 6/1/89"

    self < 1
        ifTrue:[ ^self error: self class name,
            ' gammaLn invalid on values < 1' ].
    ^((self - 1) factorial asFloat ln )! !


!SequenceableCollection methodsFor: 'enumerating' stamp: 'jpwj 7/19/2001 16:29'!
doIndex: aBlock
    "Answer the receiver.  For each element in the receiver, evaluate aBlock with the argruments: element and index  Added for MathPack "

	"Senders should be replaced with withIndexDo:
	Temporary until correction is made to MathPack."
 
    | index size |
    index := 1.
    size := self size.
    [index > size]
        whileFalse: [
            aBlock value: (self at: index) value: index.
            index := index + 1]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'jpwj 7/19/2001 16:29'!
findFirst: aBlock ifNone: noneBlock
   "Answer the index of the first element of the receiver that
    causes aBlock to evaluate to true (with that element as the argument).
    If no such element is found, answer value of noneBlock. Added by MathPack jpwj 6/1/89"

    | index size |
    size := self size.
    index := 1.
    [index <= size]
        whileTrue: [
            (aBlock value: (self at: index) )
                ifTrue: [^index].
            index := index + 1].
    ^noneBlock value! !


!String methodsFor: 'converting' stamp: 'jpwj 7/21/2001 11:10'!
asFloat
"Answer the Float value of the decimal number contained in self.
 Easily converted to answer a Fraction for the decimal number by
 removing references to asFloat   MathPack jpwj 6/1/89"
    | ePosition  whole fraction exponent period string sign |
    string := self asLowercase trimBlanks.
    ePosition := string indexOf: $e.
    ePosition > 0
        ifTrue:[
          exponent := string copyFrom: (ePosition + 1) to: string size.
          exponent := exponent asInteger.
          string := string copyFrom: 1 to: (ePosition - 1)]
        ifFalse:[ exponent := 0 ].
    period := string indexOf: $..

    period = 0
        ifTrue:[ "no period, convert to Integer then Float"
            whole := (string asInteger) asFloat.
            exponent ~= 0
                ifTrue:[ whole := whole * (10 raisedToInteger: exponent)].
            ^whole ].
    period = string size
        ifTrue:[ "no digits follow the period"
            string at: period put: $  .
            string := string trimBlanks.
            whole := (string asInteger) asFloat.
            exponent ~= 0
                ifTrue:[ whole := whole * (10 raisedToInteger: exponent)].
            ^whole ].
        "if sign is present, remove it and set sign = true/false.
         allows for the case -0.123"
    sign := false.
   ( (string at: 1) = $+ )
      ifTrue: [
        sign := false.
        string at: 1 put: $  .
        string := string trimBlanks.].
   ( (string at: 1) = $- )
      ifTrue: [
        sign := true.
        string at: 1 put: $  .
        string := string trimBlanks.].
    period := string indexOf: $..
    period = 1
        ifTrue:[ whole := 0 ]
        ifFalse:[ whole := string copyFrom: 1 to: (period - 1).
            whole := whole asInteger asFloat.].
    fraction := string copyFrom: (period + 1) to: string size.
    fraction := fraction asInteger asFloat
                    / (10 raisedToInteger: (string size - period)).
    whole := whole + fraction.
    exponent ~= 0
        ifTrue:[ whole := whole * (10 raisedToInteger: exponent) ].
    sign
      ifTrue:[ whole := whole negated ].
    ^whole! !

!String methodsFor: 'printing' stamp: 'jpwj 7/19/2001 16:19'!
printOn: aStream rightWidth: width
   "Put the character represention of self on the stream
    right justified in the width given padded on the left with blanks.
    self is truncated on right if width < self size. Added for MathPack jpwj 6/1/89"

    aStream nextPutAll: (self printRightWidth: width)! !

!String methodsFor: 'printing' stamp: 'jpwj 7/19/2001 16:21'!
printOn: aStream width: width
   "Put the character represention of self on the stream
    left justified in the width given padded on the right with blanks.
    self is truncated on right if width < self size.  Added for MathPack jpwj 6/1/89"

    aStream nextPutAll: (self printWidth: width)! !

!String methodsFor: 'printing' stamp: 'jpwj 7/19/2001 16:19'!
printRightWidth: width
   "Answer String of size width with ASCII representation of receiver
    added right justified and padded on the left with blanks. Added for MathPack jpwj 6/1/89"
    | string padd |
    padd := width - self size.
    padd < 1  "self size >= width, return truncated portion"
        ifTrue:[ ^self copyFrom: 1 to: width ].
    string := String new: padd.
    ^string,self! !

!String methodsFor: 'printing' stamp: 'jpwj 7/21/2001 13:31'!
printWidth: width
   "Answer String of size width with ASCII representation of receiver
    added left justified and padded on the right with blanks. Added for Mathpack jpwj 6/1/89"
    | string replaceSize |
    string := String new: width.
    self size < width
        ifTrue: [ replaceSize := self size ]
        ifFalse:[ replaceSize := width ].
    ^string replaceFrom: 1 to: replaceSize with: self startingAt: 1! !


!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/20/2001 09:08'!
* aScalarVectorOrMatrix
   "Answer new instance equal to self * aScalarVectorOrMatrix
    Operation performs element wise multiplication for
    a scalar or another vector of correct size.
    NOTE: When aScalarVectorOrMatrix is a Matrix,
        self MUST contain ONLY scalar values (no nil values).
    Method times: does operation on contents of self
                  for aScalar or aVector.
    If aScalarVectorOrMatrix is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
    |  answer |
 "Is aScalarVectorOrMatrix a scalar?"
    aScalarVectorOrMatrix isScalar
      ifTrue:[
        answer := Vector new: self size.
        ^self timesScalar: aScalarVectorOrMatrix
                   result: answer  ].
 "Is aScalarVectorOrMatrix a Vector ?"
    (aScalarVectorOrMatrix isKindOf: Vector )
      ifTrue:[ "Same size?"
        (self size = aScalarVectorOrMatrix size )
          ifFalse: [
            ^self error: self class name,' *  wrong size Vector' ].
        answer := Vector new: self size.
        ^self timesVector: aScalarVectorOrMatrix
              result: answer  ].
 "Is aScalarVectorOrMatrix a Matrix ?"
    (aScalarVectorOrMatrix isKindOf: Matrices2Dimensional )
      ifTrue:[ "self is considered a 1 @ self size matrix"
         ^self timesMatrix: aScalarVectorOrMatrix ].
    ^self error: self class name,' * invalid object'! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 16:35'!
+ aScalarOrVector
   "Answer a new instance equal to self + aScalarOrVector
    where aScalarOrVector is either a scalar or another vector.
    element wise addition
    Method plus: does operation on contents of self.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
    | newVector |
 "Is aScalarOrVector a scalar?"
    aScalarOrVector isScalar
      ifTrue: [
        newVector := Vector new: self size.
        ^self plusScalar: aScalarOrVector result: newVector ].
 "Is aScalarOrVector an Vector ?"
    (aScalarOrVector isKindOf: Vector )
      ifFalse: [
        ^self error: self class name,' + invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' + wrong size Vector' ].
    newVector := Vector new: self size.
    ^self plusVector: aScalarOrVector result: newVector! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/20/2001 09:05'!
- aScalarOrVector
   "Answer a new instance of self - aScalarOrVector
    where aScalarOrVector is either a scalar or vector.
    element wise subtraction.
    Method minus: does operation on contents of self.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
    | newVector |
 "Is aScalarOrVector a scalar?"
    aScalarOrVector isScalar
      ifTrue: [
        newVector := Vector new: self size.
        ^self minusScalar: aScalarOrVector result: newVector ].
 "Is aScalarOrVector an Vector ?"
    (aScalarOrVector isKindOf: Vector)
      ifFalse: [
        ^self error: self class name,' - invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' -  wrong size Vector' ].
    newVector := Vector new: self size.
    ^self minusVector: aScalarOrVector result: newVector! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/20/2001 08:43'!
/ aScalarOrVector
   "Answer a new instance of self / aScalarOrVector
    where aScalarOrVector is either a scalar or another vector.
    element wise division.
    Method divide: does operation on contents of self.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
    | newVector |
 "Is aScalarOrVector a scalar?"
    aScalarOrVector isScalar
      ifTrue: [
        newVector := Vector new: self size.
        ^self divideByScalar: aScalarOrVector result: newVector].
 "Is aScalarOrVector an Vector ?"
    aScalarOrVector class = Vector
      ifFalse: [
        ^self error: self class name,' / invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' /  wrong size Vector' ].
    newVector := Vector new: self size.
    ^self divideByVector: aScalarOrVector result: newVector
! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 16:30'!
crossProduct
   "Answer a Matrix which contains the cross product of self by self.
    self MUST contain NO nil elements. See method #crossProduct:
    for more detailed explanation. "

    ^self crossProduct: self! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/20/2001 08:46'!
dotProduct
   "Answer 0 if noElements
    Else answer the sum of the elements squared.
    self MUST contain NO nil elements."
    ^(self dotProduct: self)! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/19/2001 16:35'!
negated
   "Answer a new instance of self with the nonNil values negated."
    ^self copy contentsNegated! !

!Vector methodsFor: 'arithmetic' stamp: 'jpwj 7/20/2001 09:02'!
reciprocal
   "Answer an new instance self with the reciprocal of the nonNil values.
    **NO precaution is taken to detect reciprocals of zero values**"
    ^self copy contentsReciprocal! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:53'!
divide: aScalarOrVector
   "replace contents of self with self / aScalarOrVector
    where aScalarOrVector is either a scalar or another vector.
    element wise division.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
 "Is aScalarOrVector a scalar?"
    aScalarOrVector isScalar
      ifTrue:[
        ^self divideByScalar: aScalarOrVector result: self ].
 "Is aScalarOrVector an Vector ?"
    aScalarOrVector class = Vector
      ifFalse: [
        ^self error: self class name,' / invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' /  wrong size Vector' ].
    ^self divideByVector: aScalarOrVector result: self! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:58'!
divideByScalar: aScalar result: answerVector
   "Private - divide each non-nil element in self by aScalar
    and place answer in answerVector. "
    self doIndex: [ :x :index |
        x isNil
          ifFalse: [
            answerVector at: index put: (x / aScalar ).].].
    ^answerVector! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 09:07'!
divideByVector: aVector result: answerVector
   "Private - element wise division of elements
    in self and aVector with result place in answerVector.
    No action is taken if (self at: index) isNil.
    (self at: index) is set to nil when aVector at: index isNil"
    self doIndex: [ :x :index |
        x isNil ifFalse: [
          (aVector at: index) isNil
            ifTrue:  [ answerVector at: index put: nil ]
            ifFalse: [
              answerVector at: index put: (x / (aVector at: index)).].].].
    ^answerVector! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:53'!
minus: aScalarOrVector
   "changes the elements in self to self - aScalarOrVector
    where aScalarOrVector is either a scalar or another vector.
    element wise subtraction.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
 "Is anObject a scalar?"
    aScalarOrVector isScalar
      ifTrue: [
        ^self minusScalar: aScalarOrVector result: self].
 "Is anObject a Vector ?"
    (aScalarOrVector isKindOf: Vector )
      ifFalse: [
        ^self error: self class name,' - invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' -  wrong size Vector' ].
    ^self minusVector: aScalarOrVector result: self! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:58'!
minusScalar: aScalar result: answerVector
   "Private - subtract a scalar from each non nil element in self
    placing the result in a answerVector "
    self doIndex: [ :x :index |
        x isNil
            ifFalse: [ answerVector at: index put: (x - aScalar) ].].
    ^answerVector! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:56'!
plus: aScalarOrVector
   "changes the elements is self to self + aScalarOrVector.
    element wise addition.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."
 "Is aScalarOrVector a scalar?"
    aScalarOrVector isScalar
      ifTrue: [
        ^self plusScalar: aScalarOrVector result: self ].
  "Is aScalarOrVector an Vector ?"
    aScalarOrVector class = Vector
      ifFalse: [
        ^self error: self class name,' + invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' +  wrong size Vector' ].
    ^self plusVector: aScalarOrVector result: self! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:46'!
plusScalar: aScalar result: answerVector
   "Private - add a scalar to each non nil element in self
    placing the result in a answerVector "
    self doIndex: [ :x :index |
        x isNil
            ifFalse: [ answerVector at: index put: (x + aScalar) ].].
    ^answerVector! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/19/2001 16:36'!
plusVector: aVector result: answerVector
   "Private - for each element in self add the corresponding
    element from aVector and place the result in a answerVector
    No action is taken if (self at: index) isNil.
    (self at: index) is set to nil when answerVector at: index isNil"
    self doIndex: [ :x :index |
        x isNil ifFalse: [
          (aVector at: index) isNil
            ifTrue:  [ answerVector at: index put: nil ]
            ifFalse: [
              answerVector at: index put: (x + (aVector at: index)).].].].
    ^answerVector! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 09:04'!
plusVector: aVector times: aScalar 
   "Answer self after adding to each element in self the 
    corresponding elements in aVector times aScalar.  i.e. 
    e = e + (a * s)  where e is (self at: i), a is (aVector at: i) 
    and s is aScalar for i = 1 to self size. 
       To subtract (a * s) instead of adding use aScalar negated." 
    | index | 
      "Is aScalar a scalar?" 
    aScalar isScalar 
      ifFalse: [ 
        ^self error: self class name,' not aScalar.' ]. 
      "Is aVector the correct size?" 
    (self size = aVector size ) 
      ifFalse: [ 
        ^self error: self class name,' aVector is wrong size.' ]. 
    index := 0. 
    self do: [ :x | 
        index := index + 1. 
        (x isNil or:[ (aVector at: index) isNil]) 
          ifFalse: [ 
            self at: index 
                put: (x + ((aVector at: index) * aScalar)).].  ]. 
    ^self! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:57'!
times: aScalarOrVector
   "change contents of self to self * aScalarOrVector
    Operation performs element wise multiplication for
    a scalar or another vector of correct size.
    If aScalarOrVector is a scalar or vector, then nil elements
    are allowed and the (answerVector at: anIndex) is set to nil if the
    corresponding element in self or vector is nil."

 "Is aScalarOrVector a scalar?"
    aScalarOrVector isScalar
      ifTrue: [
        ^self timesScalar: aScalarOrVector result: self  ].
 "Is anObject an Vector ?"
    (aScalarOrVector isKindOf: Vector)
      ifFalse: [
        ^self error: self class name,' * invalid object' ].
   "Same size?"
    (self size = aScalarOrVector size )
      ifFalse: [
        ^self error: self class name,' *  wrong size Vector' ].
    ^self timesVector: aScalarOrVector result: self! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/19/2001 16:35'!
timesMatrix: aMatrix
   "Private - answer the appropiate object(scalar or vector)
    resulting from self * aMatrix.  self treated as row vector
    i.e. matrix with dimensions of 1 @ self size
    and self MUST contain no nil elements."
    ^(self asMatrix: 1 @ self size) timesMatrix: aMatrix! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 09:05'!
timesScalar: aScalar result: answerVector
   "Private - multiple each non-nil element in self by aScalar
    and place answer in answerVector. "
    self doIndex: [ :x :index |
        x isNil
          ifFalse: [
            answerVector at: index put: (x * aScalar ).].].
    ^answerVector! !

!Vector methodsFor: 'arithmetic-private' stamp: 'jpwj 7/20/2001 08:52'!
timesVector: aVector result: answerVector
   "Private - element wise multiplication of elements
    in self and aVector with result place in answerVector.
    No action is taken if (self at: index) isNil.
    (self at: index) is set to nil when aVector at: index isNil"
    self doIndex: [ :x :index |
        x isNil ifFalse: [
          (aVector at: index) isNil
            ifTrue:  [ answerVector at: index put: nil ]
            ifFalse: [
              answerVector at: index put: (x * (aVector at: index)).].].].
    ^answerVector! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/20/2001 08:53'!
contentsFloat
   "Answer self after converting nonNil values to class Float.
    nonNil values are ASSUMED to belong to class Number or ComplexNumber.
    Used to enhance speed of calculations."
    self doIndex:[ :x :index |
      x isNil
        ifFalse:[ self at: index put: x asFloat ]. ]! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/20/2001 08:59'!
contentsForceToZero: ifLessThanEpsilon
   "Answer self after replacing all nonNil elements that
    have an absolute value less than epsilon with zero.
    nonNil values are ASSUMED to belong to class Number or ComplexNumber."
    self doIndex:[ :x :index |
      x isNil
        ifFalse:[ self at: index put: (x forceToZero: ifLessThanEpsilon)  ]. ].
! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/20/2001 09:08'!
contentsNegated
   "Answer self after negating nonNil values.
    nonNil values are ASSUMED to belong to class Number or ComplexNumber."
    self doIndex:[ :x :index |
      x isNil
        ifFalse:[ self at: index put: x negated ]. ]! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/20/2001 09:09'!
contentsReciprocal
   "Answer self with reciprocal of nonNil values.
    nonNil values are ASSUMED to belong to class Number or ComplexNumber.

    *****  NO check is made for division by Zero. *****
"
    self doIndex:[ :x :index |
      x isNil
        ifFalse:[ self at: index put: x reciprocal ]. ]! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/19/2001 16:32'!
convertNilSymbols 
   "Answer the receiver with all nil symbols converted 
    to the UndefinedObject nil." 
    | nilSymbol | 
    nilSymbol := 'nil' asSymbol. 
    1 to: self size do:[ :i | 
        (self at: i) = nilSymbol 
            ifTrue:[ self at: i put: nil]. ]! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/20/2001 09:07'!
formComplexNumbers
   "Answer a new instance with complex numbers formed
    from the number preceeding and following the symbol 'i:'.
    e.g. #( 1i:6 2i:2) asVector produces a vector
           with six elements 1,i:,6,2,i:,2.
    #( 1i:6 2i:2) asVector formComplexNumbers answers a vector
    of two elements, the complex numbers 1i:6 and 2i:2. "
    | answer iSymbol index1 index2 size x |
    iSymbol := 'i:' asSymbol.
    size := self size.
    answer := Vector new: size.
    index1 := index2 := 0.
    [ size > index1 ] whileTrue:[
        index1 := index1 + 1.
        x := self at: index1.
        x = iSymbol
          ifFalse:[
            index2 := index2 + 1.
            answer at: index2 put: x ]
          ifTrue:[
            answer at: index2
                  put:( (answer at: index2) i: (self at: index1 + 1)).
            index1 := index1 + 1. ].].
    ^answer copyFrom: 1 to: index2! !

!Vector methodsFor: 'converting' stamp: 'jpwj 7/19/2001 16:31'!
formComplexNumbersPairwise
   "Answer a new instance with complex numbers formed pairwise
    from the elements is self.
    e.g. #(-1 0  -9 10) asVector formComplexNumbersPairwise
    answer a vector with two elements, the complex numbers
    -1i:0 and -9i:10 "
    | answer index1 index2 size |
    size := self size.
    answer := Vector new: size / 2.
    index1 := 1.
    index2 := 0.
    [ size > index1 ] whileTrue:[
        index2 := index2 + 1.
        answer at: index2
             put:( (self at: index1) i: (self at: index1 + 1)).
        index1 := index1 + 2.].
    ^answer! !

!Vector methodsFor: 'math functions' stamp: 'jpwj 7/20/2001 08:44'!
countWithMinimumAndMaximum
   "answer an Vector with count, minimum and maximum values
        or nil if empty.
    ASSUMES all elements are of the same class, Number, Character,
    String, Symbol, etc. but not of class ComplexNumber"
    | count minimum maximum index |
    index := self findFirst: [ :x | x ~= nil ] ifNone: [^nil].
    index isNil ifTrue: [ ^index ].
    minimum := self at: index.
    maximum := self at: index.
    count := 0.
    index to: self size do:[ :i |
      (self at: i) isNil
        ifFalse:[
            count := count + 1.
            minimum := (self at: i) min: minimum.
            maximum := (self at: i) max: maximum ].].
    ^Vector with: count with: minimum with: maximum
! !

!Vector methodsFor: 'math functions' stamp: 'jpwj 7/20/2001 08:45'!
forceToZero: ifLessThanEpsilon
   "Answer an new instance of self after replacing all nonNil elements
    that have an absolute value less than epsilon with zero.
    nonNil values are ASSUMED to belong to class Number or ComplexNumber."
    ^self copy contentsForceToZero: ifLessThanEpsilon! !

!Vector methodsFor: 'math functions' stamp: 'jpwj 7/20/2001 09:02'!
scaleFrom: startValue  to: endValue inset: inset 
    " Scale self to fit between startValue and endValue 
      adjusted by inset.  e.g. scaleFrom: 0 to: 480 inset: 10 
      would adjust values to fall into the range 10 to: 470 " 
      | start end scale temp minval | 
    ( startValue < endValue ) 
      ifTrue: [ 
        start := (startValue + inset ) asFloat. 
        end   := (endValue - inset ) asFloat.] 
      ifFalse: [ 
        start := (startValue - inset ) asFloat. 
        end   := (endValue + inset ) asFloat.]. 
    minval := (self min) asFloat. 
    scale := (end - start ) / (( self max) - minval). 
     1 to: self size do:[ :i | 
      temp := ( self at: i ). 
      (temp = nil) 
        ifFalse: [ 
          (temp - minval ) = 0 
            ifTrue: [ self at: i put: (start rounded).] 
            ifFalse:[ self at: i 
         put: ( ( ( (temp - minval)* scale)+start) rounded)]. ]. ].
! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/19/2001 16:34'!
absoluteDeviation 
   "answer the absoluteDeviation of the nonNil numerical values. 
    Report an error if the number of nonNil elements < 2." 
    | absoluteDeviation mean count | 
    count := (self checkCount:'absoluteDeviation') asFloat. 
    mean := self mean asFloat. 
    absoluteDeviation := 0.0. 
    self do:[ :x | 
      x isNil 
        ifFalse:[ 
          absoluteDeviation := absoluteDeviation + ((x - mean) abs). ]. ]. 
    ^absoluteDeviation / count! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:58'!
count 
   "answer the number of nonNil values in the receiver" 
    | count | 
    count := 0. 
    self do:[ :x | 
      x isNil 
        ifFalse: [ count := count + 1 ]. ]. 
    ^count! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:54'!
countMeanVariance
   "Answer an vector with count, mean, variance
        for the nonNil numerical values of self.
    Report an error if the number of nonNil elements < 2."
    | variance count mean sumOfElements sumOfElementsSquared vector |
    count := (self checkCount:'countMeanVariance') asFloat.
    vector := self countWithSumSumSquares.
    sumOfElements := vector at: 2.
    sumOfElementsSquared := vector at: 3.
    mean := sumOfElements / count.
    variance := ( (sumOfElementsSquared - ( count * (mean * mean)) )
                     / (count - 1) ).
    ^Vector with: count with: mean with: variance! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:59'!
countWithSumOfElements
   "Answer an vector with count and sum of the
        nonNil numerical values of self."
    |  count  sumOfElements |
    sumOfElements := 0.  count := 0.
    self do:[ :x |
      x isNil
        ifFalse:[
          sumOfElements := sumOfElements + x.
          count := count + 1. ]. ].
    ^Vector with: count with: sumOfElements! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 09:00'!
countWithSumSumSquares
   "Answer an vector with
        count, sum of elements, sum of elements squared.
    nil elements are ignored. "
    |  count  sumOfElements sumOfElementsSquared |
    count := 0.  sumOfElements := 0.  sumOfElementsSquared := 0.
    self do:[ :x |
      x isNil
        ifFalse:[
          sumOfElements := sumOfElements + x.
          sumOfElementsSquared := sumOfElementsSquared + (x * x).
          count := count + 1. ]. ].
    ^Vector with: count with: sumOfElements with: sumOfElementsSquared
! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:54'!
kurtosis 
   "answer the kurtosis of the nonNil numerical values. 
    Report an error if the number of nonNil elements < 2. 
    answer nil if variance equals 0.0 i.e. can not compute kurtosis " 
    | count variance mean kurtosis temp | 
    count := (self checkCount:'kurtosis') asFloat. 
    kurtosis := 0.0. 
    ((variance := self variance) = 0.0 ) ifTrue:[^nil ]. 
    mean := self mean. 
    self do:[ :x | 
      x isNil 
        ifFalse:[ 
          temp := x - mean. 
          temp := temp * temp. 
          kurtosis := kurtosis + (temp * temp). ]. ]. 
    kurtosis := (kurtosis / (count * (variance * variance))) - 3.0. 
    ^kurtosis! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:51'!
maximum 
   "answer the maximum value or nil if empty 
    ASSUMES all elements are of the same class, Number, Character, 
    String, Symbol, etc. but not of ComplexNumber" 
    | maximum index | 
    index := self findFirst: [ :x | x ~= nil ] ifNone: [^nil]. 
    index isNil ifTrue: [ ^index ]. 
    maximum := self at: index. 
    index to: self size do:[ :i | 
      (self at: i) isNil 
        ifFalse:[ maximum := (self at: i) max: maximum ].]. 
    ^maximum! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:49'!
mean 
   "answer the mean (average) of the nonNil numerical values. 
    Report an error if the number of nonNil elements < 2." 
    | count sumOfElements | 
    count := (self checkCount:'mean'). 
    sumOfElements := 0. 
    self do:[ :x | 
      x isNil 
        ifFalse:[ 
          sumOfElements := sumOfElements + x. ]. ]. 
    ^sumOfElements / count! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/19/2001 16:37'!
median 
   "Answer the median value of the nonNil values in self. 
    Answer an error if the number of nonNil values < 2." 
    | index count sortedVector halfCount | 
    count := self checkCount: 'median'. 
    self size = count 
      ifTrue:  [ sortedVector := self asSortedCollection ] 
      ifFalse: [ "extracted non-nil values" 
        sortedVector := Vector new: count. 
        index := 0. 
        self do:[ :x | 
            x isNil 
                ifFalse:[ 
                    index := index + 1. 
                    sortedVector at: index put: x ]. ]. 
        sortedVector := sortedVector asSortedCollection. ]. 
    halfCount := count // 2. 
    ( 2 * halfCount = count ) 
        ifTrue: [ ^( ((sortedVector at: halfCount)  + 
            (sortedVector at: (halfCount + 1) ) / 2) ) ] 
        ifFalse:[ ^(sortedVector at: (halfCount + 1) ) ]! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:49'!
minimum 
   "answer the minimum value or nil if empty 
    ASSUMES all elements are of the same class, Number, Character, 
    String, Symbol, etc. but not of class ComplexNumber" 
    | minimum index | 
    index := self findFirst: [ :x | x ~= nil ] ifNone: [^nil]. 
    index isNil ifTrue: [ ^index ]. 
    minimum := self at: index. 
    index to: self size do:[ :i | 
      (self at: i) isNil 
        ifFalse:[ minimum := (self at: i) min: minimum ].]. 
    ^minimum! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:52'!
moments
   "Answer an Dictionary containing the statistical description of self.
    self ASSUMED to contain numerical values.  nil values are allowed.
    The keys in the dictionary are:
        count, minimum, maximum, mean, median, variance,
        standardDeviation, absoluteDeviation, skewness, kurtosis.
    nil is return if count of nonNil elements < 1.
    nil is stored for non computed keys.
    Answer an error if the number of nonNil elements < 2."
    | count moments mean temp xSquared variance standardDeviation
      absoluteDeviation skewness kurtosis tempVector |
    tempVector := self countWithMinimumAndMaximum.
    count := (tempVector at: 1).
    count isNil
        ifTrue:[ ^nil].
    moments := Dictionary new.
    moments at: 'count' put: (count := count asFloat).
    moments at: 'minimum' put: (tempVector at: 2) asFloat.
    moments at: 'maximum' put: (tempVector at: 3) asFloat.
    moments at: 'mean' put: nil.
    moments at: 'median' put: nil.
    moments at: 'variance' put: nil.
    moments at: 'standardDeviation' put: nil.
    moments at: 'absoluteDeviation' put: nil.
    moments at: 'skewness' put: nil.
    moments at: 'kurtosis' put: nil.
    count < 2
        ifTrue:[ ^moments ].
    moments at: 'mean' put: (mean := self mean asFloat).
    moments at: 'median' put: (self median asFloat).
    variance := 0.0.  absoluteDeviation := 0.0.
    skewness := 0.0.  kurtosis := 0.0.
    self do:[ :x |
      x isNil
        ifFalse:[
          temp := x - mean.
          absoluteDeviation := absoluteDeviation + (temp abs).
          xSquared := temp * temp.
          variance := variance + xSquared.
          skewness := skewness + ( xSquared * temp ).
          kurtosis := kurtosis + ( xSquared * xSquared ). ]. ].
    moments at: 'variance'
          put: (variance := (variance / (count - 1.0)) ).
    moments at: 'standardDeviation'
           put: (standardDeviation := (variance sqrt)).
    moments at: 'absoluteDeviation'
           put: (absoluteDeviation := (absoluteDeviation / count)).
    variance = 0.0
      ifTrue:[ ^moments ].
    skewness := skewness / (count * (standardDeviation raisedTo: 3)).
    kurtosis := (kurtosis / (count * (variance * variance))) - 3.0.
    moments at:  'skewness' put: skewness.
    moments at: 'kurtosis' put: kurtosis.
    ^moments! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 09:08'!
skewness 
   "answer the skewness of the nonNil numerical values. 
    Report an error if the number of nonNil elements < 2. 
    answer nil if variance equals 0.0 i.e. can not compute skewness" 
    | count variance mean standardDeviation skewness temp | 
    count := (self checkCount:'skewness') asFloat. 
    skewness := 0.0. 
    ((variance := self variance) = 0.0 ) ifTrue:[^nil ]. 
    standardDeviation := variance sqrt. 
    mean := self mean asFloat. 
    self do:[ :x | 
      x isNil 
        ifFalse:[ 
          temp := x - mean. 
          skewness := skewness + (temp * temp * temp). ]. ]. 
    skewness := skewness / (count * (standardDeviation raisedTo: 3)). 
    ^skewness! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/20/2001 08:50'!
standardDeviation 
   "compute the standard deviation of non-nil numeric values" 
    self checkCount:'standardDeviation'. 
    ^self variance sqrt! !

!Vector methodsFor: 'moments' stamp: 'jpwj 7/19/2001 16:34'!
variance
   "answer the variance of the nonNil numerical values.
    Report an error if the number of nonNil elements < 2."
    | variance count mean sumOfElements sumOfElementsSquared vector|
    count := (self checkCount:'variance') asFloat.
    vector := self countWithSumSumSquares.
    sumOfElements := vector at: 2.
    sumOfElementsSquared := vector at: 3.
    mean := sumOfElements / count.
    variance := ( (sumOfElementsSquared - ( count * (mean * mean)) )
                     / (count - 1) ).
    ^variance! !

!Vector methodsFor: 'printing' stamp: 'jpwj 7/19/2001 16:37'!
printOn: aStream width: width decimal: places
   "Print vector of numeric values, right justified in width given
    with the number of specified decimal places."
    self do:[ :x |
      x isNil
        ifTrue: [ 'nil' printOn: aStream rightWidth: width ]
        ifFalse:[ x printOn: aStream width: width decimal: places ]. ].
    aStream cr! !

!Vector methodsFor: 'private' stamp: 'jpwj 7/19/2001 16:36'!
checkCount: aString
   "Answer the count of the nonNil elements or Answer an error
    when count < 2.  A number of methods should not be used unless
    the number of non-nil values is 2 or greater.
    aString should be the method name requesting the count check."
    | count |
    count := self count.
    count < 2 ifTrue:[
       ^self error: aString,' requires at least 2 values.' ].
    ^count! !

!Vector methodsFor: 'private' stamp: 'jpwj 7/20/2001 08:48'!
minusVector: aVector result: answerVector
   "Private - for each element in self substract the corresponding
    element from aVector and place the result in a newVector
    No action is taken if (self at: index) isNil.
    (self at: index) is set to nil when aVector at: index isNil"
    self doIndex: [ :x :index |
        x isNil ifFalse: [
          (aVector at: index) isNil
            ifTrue:  [ answerVector at: index put: nil ]
            ifFalse: [
              answerVector at: index put: (x - (aVector at: index)).].].].
    ^answerVector! !

